package com.esk.tree;

import static org.objectweb.asm.tree.AbstractInsnNode.*;

import java.util.Arrays;

import org.objectweb.asm.Opcodes;
import org.objectweb.asm.tree.*;

import com.esk.util.*;

public class Node extends Tree<Node> implements Opcodes {
  protected NodeTree t;
  protected AbstractInsnNode ai;
  public final int c, p;
  protected int r;
  public boolean s, e, h;

  public Node(NodeTree t, AbstractInsnNode ai, int c, int p) {
    this.t = t;
    this.ai = ai;
    this.c = c;
    this.p = r = p;
  }

  public void accept(Visitor v) {
    switch(ai.getType()) {
      case INSN:
        v.visit(this);
        break;
      case INT_INSN:
        v.visitInt(this);
        break;
      case VAR_INSN:
        v.visitVar(this);
        break;
      case TYPE_INSN:
        v.visitType(this);
        break;
      case FIELD_INSN:
        v.visitField((InvocationNode)this);
        break;
      case METHOD_INSN:
        v.visitMethod((InvocationNode)this);
        break;
      case JUMP_INSN:
        throw new RuntimeException();
      case LABEL:
        v.visitLabel(this);
        break;
      case LDC_INSN:
        v.visitLdc(this);
        break;
      case IINC_INSN:
        v.visitIinc(this);
        break;
      case TABLESWITCH_INSN:
        v.visitTableSwitch(this);
        break;
      case LOOKUPSWITCH_INSN:
        v.visitLookupSwitch(this);
        break;
      case MULTIANEWARRAY_INSN:
        v.visitMultiANewArray(this);
        break;
      case FRAME:
        v.visitFrame(this);
        break;
      case LINE:
        v.visitLine(this);
        break;
      default:
        throw new RuntimeException();
    }
  }

  public void splice(Node n) {
    int i = indexOf(n);
    remove(i);
    while(n.has())
      add(i, n.popTail());
  }

  public void flip() {
    List<Node> ns = new List<Node>();
    Node pr = previous();
    if(pr instanceof JumpNode && contains(((JumpNode)pr).target().resolve())) {
      JumpNode jn = (JumpNode)pr;
      Target t = jn.target();
      if(t.previous().op() == GOTO) {
        if(t.owner().indexOf(t) != 2)
          throw new RuntimeException();
        for(int i = 0; i < 4; i++)
          ns.add(popHead());
      }
    }
    List<Node> p = new List<Node>();
    p.add(producing());
    for(int i = size() - 1; i >= 0; i--) {
      Node n = get(i);
      if(n.p > 0) {
        if(i > 0) {
          pr = get(i - 1);
          if(pr instanceof JumpNode
              && n.contains(((JumpNode)pr).target().resolve())) {
            p.remove(n);
            continue;
          }
        }
        set(i, p.popHead());
      }
    }
    if(ns.size() > 0)
      add(ns);
  }

  public void setInsn(AbstractInsnNode ai) {
    this.ai = ai;
  }

  public boolean delete() {
    return owner().remove(this);
  }

  @SuppressWarnings("unchecked")
  public void deleteTryCatch() {
    if(!h)
      throw new IllegalArgumentException();
    MethodNode mn = t.method();
    java.util.List<TryCatchBlockNode> tcbs = mn.tryCatchBlocks;
    List<TryCatchBlockNode> tbd = new List<TryCatchBlockNode>();
    for(TryCatchBlockNode tcb : tcbs) {
      AbstractInsnNode h = Util.nextReal(tcb.handler);
      if(h == ai)
        tbd.add(tcb);
    }
    for(TryCatchBlockNode tcb : tbd)
      tcbs.remove(tcb);
  }

  public MethodNode method() {
    return t.method();
  }

  public NodeTree tree() {
    return t;
  }

  public Node[] producing() {
    int len = size();
    Node[] ns = new Node[len];
    int i = 0;
    for(Node n : this)
      if(n.p > 0)
        ns[i++] = n;
    return Arrays.copyOf(ns, i);
  }

  public Node previous() {
    Node o = owner();
    int i = o.indexOf(this);
    if(i - 1 == -1)
      return null;
    return o.get(i - 1);
  }

  public Node next() {
    Node o = owner();
    int i = o.indexOf(this);
    if(i + 1 == o.size())
      return null;
    return o.get(i + 1);
  }

  public AbstractInsnNode[] collapse() {
    AbstractInsnNode[] ais = new AbstractInsnNode[totalSize()];
    int i = 0;
    for(Node n : this) {
      AbstractInsnNode[] _ais = n.collapse();
      System.arraycopy(_ais, 0, ais, i, _ais.length);
      i += _ais.length;
    }
    if(ais.length - i != 1)
      throw new RuntimeException();
    ais[i] = ai;
    return ais;
  }

  public AbstractInsnNode insn() {
    return ai;
  }

  public String toString() {
    return toString(1);
  }

  public String toString(int i) {
    StringBuilder sb = new StringBuilder();
    sb.append(Util.toString(ai));
    for(Node n : this) {
      sb.append('\n');
      for(int j = 0; j < i; j++)
        sb.append(' ');
      sb.append(n.toString(i + 1));
    }
    return sb.toString();
  }

  public int totalSize() {
    int s = 1;
    for(Node n : this)
      s += n.totalSize();
    return s;
  }

  public int op() {
    return ai.getOpcode();
  }
}